
import java.net.*;
import java.io.*;
import java.nio.*;
import java.util.regex.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.text.*;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.border.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.nio.channels.*;
import java.awt.image.*;
import javax.imageio.*;
import javax.net.*;
import javax.net.ssl.*;
import java.security.*;
import java.util.jar.*;
import java.security.cert.*;
import java.math.*;

import org.jdom.*;
import org.jdom.adapters.*;
import org.jdom.filter.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.transform.*;
import org.jdom.xpath.*;

public class GameSession {
    
    public static void main(String[] args) throws Exception {
        GameSession gs = new GameSession(new Board("aboard.xml"));
        
        for (int symmetricSpace = 0; symmetricSpace < 8; symmetricSpace++) {
            System.out.println("symmetricSpace = " + symmetricSpace);
            System.out.println(gs.currentStateString(0, symmetricSpace));
        }
    }
    
    public Board board;
    public boolean gameOver;
    public Vector<Spot> explored = new Vector<Spot>();
    public Set<Spot> mistakes = new HashSet<Spot>();
    public Vector<Integer> interestingTimes = new Vector<Integer>();
    public Vector<String> annotations = new Vector<String>();
    
    public GameSession(Board board) {
        this.board = board;
        this.gameOver = false;
        guess(board.start, false);
    }
    
    public GameSession(String filename) throws Exception {
        this(u.getElement(u.loadXml(filename), "session"));
    }
    
    public GameSession(Element e) throws Exception {
        board = new Board(u.getElement(e, "board"));
        gameOver = new Boolean(e.getAttributeValue("gameOver"));
        explored = (Vector<Spot>)elementToVector(u.getElement(e, "explored"), Spot.class);
        mistakes = (Set<Spot>)elementToSet(u.getElement(e, "mistakes"), Spot.class);
        interestingTimes = (Vector<Integer>)elementToVector(u.getElement(e, "interestingTimes"), Integer.class);
        annotations = (Vector<String>)elementToVector(u.getElement(e, "annotations"), String.class);
    }
    
    public void save(String filename) throws Exception {
        u.save(new File(filename), u.makeDocument(toElement()));
    }
    
    public Element toElement() {
        Element e = new Element("session");
        e.setAttribute("gameOver", "" + gameOver);
        e.addContent(board.toElement());
        e.addContent(toElement("explored", explored));
        e.addContent(toElement("mistakes", mistakes));
        e.addContent(toElement("interestingTimes", interestingTimes));
        e.addContent(toElement("annotations", annotations));
        return e;
    }
    
    public Element toElement(String name, Collection objs) {
        Element e = new Element(name);
        for (Object o : objs) {
            Element ee = new Element("item");
            ee.setAttribute("value", "" + o);
            e.addContent(ee);
        }
        return e;
    }
    
    public Vector elementToVector(Element e, Class c) throws Exception {
        Vector<Object> v = new Vector<Object>();
        for (Object o : XPath.selectNodes(e, "item")) {
            Element ee = (Element)o;
            v.add(c.getConstructor(String.class).newInstance(ee.getAttributeValue("value")));
        }
        return v;
    }
    
    public Set elementToSet(Element e, Class c) throws Exception {
        HashSet<Object> s = new HashSet<Object>();
        for (Object o : XPath.selectNodes(e, "item")) {
            Element ee = (Element)o;
            s.add(c.getConstructor(String.class).newInstance(ee.getAttributeValue("value")));
        }
        return s;
    }
    
    public void debugPrint() {
        System.out.println("-----------------------------");
        for (int y = 0; y < board.height; y++) {
            for (int x = 0; x < board.width; x++) {
                Spot s = new Spot(x, y);
                
                if (explored(s)) {
                    if (board.mine(s)) {
                        System.out.print("M");
                    } else {
                        System.out.print("" + board.adjacentMines(s).size());
                    }
                } else {
                    System.out.print("?");
                }
            }
            System.out.println();
        }
        System.out.println("-----------------------------");
    }
    
    public boolean gameOver() {
        return gameOver;
    }
    
    public boolean explored(Spot spot) {
        return explored.indexOf(spot) != -1;
    }
    
    public boolean mistake(Spot spot) {
        return mistakes.contains(spot);
    }
    
    public boolean exploredBy(Spot spot, int byTime) {
        int time = explored.indexOf(spot);
        return time != -1 && time <= byTime;
    }
    
    public void explore(Spot spot) {
        if (board.onBoard(spot) && !explored(spot)) {
            explored.add(spot);
            if (!board.mine(spot) && board.mineCount(spot) == 0) {
                for (Spot s : adjacentUnexplored(spot)) {
                    explore(s);
                }
            }
        }
    }
    
    public boolean markInteresting() {
        int time = explored.size() - 1;
        if (interestingTimes.isEmpty() || interestingTimes.lastElement() != time) {
            interestingTimes.add(time);
            return true;
        }
        return false;
    }
    
    public void annotate(String s) {
        annotations.add(s);
    }
    
    public void baseGuess(Spot spot, boolean mine) {
        if (!board.onBoard(spot) || explored(spot))
            return;
        
        explore(spot);
        if (board.mine(spot) != mine) {
            mistakes.add(spot);
            gameOver = true;
        }
        if (explored.size() == board.squareCount()) {
            gameOver = true;
        }
    }
    
    public boolean guess(Spot spot, boolean mine) {
        baseGuess(spot, mine);
        return markInteresting();
    }
    
    public boolean guess(Vector<Spot> spots, boolean mine) {
        for (Spot spot : spots) {
            baseGuess(spot, mine);
        }
        return markInteresting();
    }
    
    public boolean guess(Vector<Spot> markThese, Vector<Spot> clickThese, int x, int y, int symmetricSpace) {
        for (Spot s : markThese) {
            baseGuess(board.toSymmetricSpace(new Spot(x + s.x, y + s.y), symmetricSpace), true);
        }
        for (Spot s : clickThese) {
            baseGuess(board.toSymmetricSpace(new Spot(x + s.x, y + s.y), symmetricSpace), false);
        }
        return markInteresting();
    }
    
    public Vector<Spot> adjacentUnexplored(Spot center) {
        Vector<Spot> v = new Vector<Spot>();
        for (Spot spot : board.adjacent(center)) {
            if (!explored(spot)) {
                v.add(spot);
            }
        }
        return v;
    }
    
    public Vector<Spot> adjacentExplored(Spot center) {
        Vector<Spot> v = new Vector<Spot>();
        for (Spot spot : board.adjacent(center)) {
            if (explored(spot)) {
                v.add(spot);
            }
        }
        return v;
    }
    
    public Vector<Spot> adjacentExploredMines(Spot center) {
        Vector<Spot> v = new Vector<Spot>();
        for (Spot spot : adjacentExplored(center)) {
            if (board.mine(spot)) {
                v.add(spot);
            }
        }
        return v;
    }
    
    public Vector<Spot> adjacentUnexploredMines(Spot center) {
        Vector<Spot> v = new Vector<Spot>();
        for (Spot spot : adjacentUnexplored(center)) {
            if (board.mine(spot)) {
                v.add(spot);
            }
        }
        return v;
    }
    
    public String currentStateString(int edgeBufferSize, int otherSpace) {
        StringBuffer buf = new StringBuffer();
        
        for (int y = -edgeBufferSize; y < board.height + edgeBufferSize; y++) {
            for (int x = -edgeBufferSize; x < board.width + edgeBufferSize; x++) {
                Spot spot = board.toSymmetricSpace(new Spot(x, y), otherSpace);
                
                if (!board.onBoard(spot)) {
                    buf.append("|");
                } else if (!explored(spot)) {
                    buf.append("?");
                } else if (board.mine(spot)) {
                    buf.append("M");
                } else {
                    buf.append("" + board.mineCount(spot));
                }
            }
            buf.append("\n");
        }
        
        return buf.toString();
    }
}
