
import java.net.*;
import java.io.*;
import java.nio.*;
import java.util.regex.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.text.*;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.border.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.nio.channels.*;
import java.awt.image.*;
import javax.imageio.*;
import javax.net.*;
import javax.net.ssl.*;
import java.security.*;
import java.util.jar.*;
import java.security.cert.*;
import java.math.*;

public class SmartSweeper extends JFrame {
    
    public static void main(String[] args) throws Exception {
        new SmartSweeper();
    }
    
    public GameSession session;
    public Board board;
    public JSlider slider;
    public JTextArea rules;
    public JTextArea annotation;
    public boolean showRect = false;
    public int rx, ry, rw, rh, symmetricSpace;
    public BoardPanel boardPanel;
    public Thread saveThread;
    public Thread ruleRunnerThread = null;
    public JFrame helpFrame = null;
    
    public SmartSweeper() throws Exception {
        super("SmartSweeper [rules.txt]    v1.1");
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                saveRules();
                System.exit(0);
            }
        });
        saveThread = new Thread(new Runnable() {
            public void run() {
                while (true) {
                    try {
                        Thread.sleep(10 * 1000);
                        saveRules();
                    } catch (Exception e) {
                    }
                }
            }
        });
        saveThread.start();
        
        getContentPane().setLayout(new BorderLayout());
        
        rules = new JTextArea();
        rules.setFont(Font.decode("Monospaced"));
        File rulesFile = new File("rules.txt");
        if (!rulesFile.exists()) {
            rulesFile = new File("rules-init.dat");
        }
        rules.setText(u.slurp(rulesFile));
        
        JPanel boardSide = new JPanel(new BorderLayout());
        
        annotation = new JTextArea();
        annotation.setFont(Font.decode("Monospaced"));
        
        JPanel temp = new JPanel(new BorderLayout());
        temp.setPreferredSize(new Dimension(300, 100));
        temp.add(new JScrollPane(annotation));
        boardSide.add(temp);
        
        JPanel center = new JPanel(new BorderLayout());
        
        slider = new JSlider(0, 0, 0);
        slider.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                updateAnnotation();
                boardPanel.repaint();
            }
        });
        temp = new JPanel(new BorderLayout());
        temp.add(slider, BorderLayout.NORTH);
        temp.add(new JLabel("hint: select slider, and use <- and -> arrow keys"));
        center.add(temp, BorderLayout.SOUTH);
        
        JPanel boardsWrapper = new JPanel(new BorderLayout());
        boardsWrapper.add(new JLabel("click to load a board"), BorderLayout.NORTH);
        JPanel boards = new JPanel();
        File[] files = new File("boards").listFiles();
        Arrays.sort(files, new Comparator<File>() {
            public int compare(File a, File b) {
                return a.getName().compareTo(b.getName());
            }
        });
        for (File f : files) {
            Matcher m = Pattern.compile("^(.*?)\\.xml$").matcher(f.getName());
            if (m.find()) {
                JButton b = new JButton(m.group(1));
                final File F = f;
                
                if (board == null) {
                    loadBoard(F);
                }
                
                b.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        loadBoard(F);
                    }
                });
                boards.add(b);
            }
        }
        boardsWrapper.add(boards);
        center.add(boardsWrapper, BorderLayout.NORTH);
        
        boardPanel = new BoardPanel();
        center.add(boardPanel);
        
        temp = new JPanel(new GridBagLayout());
        JPanel temp2 = new JPanel(new BorderLayout());
        JButton b = new JButton(new ImageIcon(ImageIO.read(new File("images/play.png"))));
        temp2.add(b, BorderLayout.NORTH);
        b.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                run();
            }
        });
        b = new JButton("< Help");
        temp2.add(b, BorderLayout.SOUTH);
        b.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                try {
                    if (helpFrame == null) {
                        JFrame f = new JFrame("A quick reference");
                        f.getContentPane().setLayout(new BorderLayout());
                        JEditorPane t = new JEditorPane("text/html", "");
                        f.getContentPane().add(new JScrollPane(t));
                        t.setText(u.slurp("help.html"));
                        f.pack();
                        helpFrame = f;
                    }
                    helpFrame.show();
                } catch (Exception ee) {
                    throw new IllegalArgumentException(ee);
                }
            }
        });
        temp.add(temp2);
        center.add(temp, BorderLayout.WEST);
        
        temp = new JPanel(new BorderLayout());
        temp.add(center, BorderLayout.NORTH);        
        boardSide.add(temp, BorderLayout.WEST);
        
        temp = new JPanel(new BorderLayout());
        temp.add(new JScrollPane(rules));
        temp.setPreferredSize(new Dimension(300, 100));
        getContentPane().add(new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
            temp, boardSide));
        
        pack();
        show();
    }
    
    public synchronized void saveRules() {
        try {
            u.save(new File("rules.txt"), rules.getText());
        } catch (Exception e) {
            JOptionPane.showMessageDialog(this, "failed to save rules.txt, you may want to copy and paste the contents in order to make sure you don't loose any information");
        }
    }
    
    public void loadBoard(File f) {
        try {
            killRuleRunnerThread();
            
            board = new Board(f);
            session = new GameSession(board);
            slider.setMinimum(0);
            slider.setMaximum(board.squareCount());
            slider.setValue(0);
            boardPanel.repaint();
        } catch (Exception e) {
        }
    }
    
    public void killRuleRunnerThread() {
        if (ruleRunnerThread != null) {
            ruleRunnerThread.stop();
            ruleRunnerThread = null;
        }
    }
    
    public void run() {
        try {
            killRuleRunnerThread();
            
            saveRules();
            
            session = new GameSession(board);
            slider.setMinimum(0);
            slider.setMaximum(board.squareCount());
            slider.setValue(0);
            boardPanel.repaint();
            
            ruleRunnerThread = new Thread(new Runnable() {
                public void run() {
                    try {
                        new RuleRunner("rules.txt", board, session, new ActionListener() {
                            public void actionPerformed(ActionEvent e) {
                                slider.setValue(session.interestingTimes.size() - 2);
                                boardPanel.repaint();
                            }
                        });
                        slider.setMaximum(session.interestingTimes.size() - 1);
                        slider.setValue(session.interestingTimes.size() - 1);
                        boardPanel.repaint();
                        slider.requestFocus();
                    } catch (Exception e) {
                        showError(e);
                    }
                    killRuleRunnerThread();
                }
            });
            ruleRunnerThread.start();
        } catch (Exception e) {
            showError(e);
            JOptionPane.showMessageDialog(null, "" + e);
        }
    }
    
    public void showError(Exception e) {
        String nice = "" + e;
        Matcher m = Pattern.compile("(?si).*?Exception:(.*)").matcher(nice);
        if (m.find()) {
            nice = m.group(1);
        }
        
        m = Pattern.compile("see \\((\\d+),\\s*(\\d+)\\)").matcher(nice);
        if (m.find()) {
            int start = new Integer(m.group(1));
            int end = new Integer(m.group(2));
            try {
                rules.select(start, end);
                rules.requestFocus();
            } catch (Exception ee) {
            }
        }
        
        JOptionPane.showMessageDialog(null, nice);
    }
    
    public void updateAnnotation() {
        showRect = false;
        try {
            String s = session.annotations.get(slider.getValue());
            Matcher m = Pattern.compile("rx = (.*?), ry = (.*?), rw = (.*?), rh = (.*?), s = (.*?)\n\n").matcher(s);
            if (m.find()) {
                showRect = true;
                rx = new Integer(m.group(1));
                ry = new Integer(m.group(2));
                rw = new Integer(m.group(3));
                rh = new Integer(m.group(4));
                symmetricSpace = new Integer(m.group(5));
                annotation.setText(s.substring(m.end(0), s.length()));
            } else {
                throw new Exception();
            }
        } catch (Exception e) {
            annotation.setText("");
        }
        if (slider.getValue() == slider.getMaximum()) {
            if (session.mistakes.size() > 0) {
                annotation.setText("game over because of a mistake\npress <- arrow\nto see the rule that caused it...");
            } else if (session.gameOver) {
                annotation.setText("mines swept!");
            } else {
                annotation.setText("...no more rules apply.");
            }
        }
    }
    
    public class BoardPanel extends JPanel {
        Map<String, BufferedImage> images = new HashMap<String, BufferedImage>();
        int tileWidth;
        int tileHeight;
        
        public BoardPanel() throws Exception {
            for (File f : new File("images").listFiles()) {
                images.put(u.trimExtension(f.getName()), ImageIO.read(f));
            }
            BufferedImage genericTile = images.get("1");
            tileWidth = genericTile.getWidth();
            tileHeight = genericTile.getHeight();
        }
    
        public Dimension getPreferredSize() {
            return new Dimension(board.width * tileWidth + (2 * tileWidth),
                board.height * tileHeight + (2 * tileWidth));
        }
        
        public String getSpotType(int x, int y) {
            Spot spot = new Spot(x, y);
            int time = Math.min(slider.getValue(), session.interestingTimes.size() - 1);
            if (time >= 0 &&
                session.exploredBy(spot, session.interestingTimes.get(time)))
            {
                if (session.mistake(spot)) {
                    if (board.mine(spot)) {
                        return "red_mine";
                    } else {
                        return "x_mine";
                    }
                } else if (board.mine(spot)) {
                    return "mine";
                } else {
                    return "" + board.mineCount(spot);
                }
            } else {
                return "unexplored";
            }
        }
        
        public void paint(Graphics g1) {
            super.paint(g1);
            Graphics2D g = (Graphics2D)g1;
            
            for (int x = 0; x < board.width; x++) {
                for (int y = 0; y < board.height; y++) {
                    g.drawImage(images.get(getSpotType(x, y)), tileWidth + x * tileWidth, tileWidth + y * tileHeight, this);
                }
            }
            
            if (showRect) {
                BufferedImage up = images.get("up");
                AffineTransform t = new AffineTransform();
                t.translate(tileWidth + rx * tileWidth + (tileWidth / 2), tileWidth + ry * tileHeight + (tileWidth / 2));
                if ((symmetricSpace & 4) != 0) {
                    t.concatenate(new AffineTransform(0, 1, 1, 0, 0, 0));
                }
                if ((symmetricSpace & 1) != 0) {
                    t.concatenate(new AffineTransform(-1, 0, 0, 1, 0, 0));
                }
                if ((symmetricSpace & 2) != 0) {
                    t.concatenate(new AffineTransform(1, 0, 0, -1, 0, 0));
                }
                t.translate(-tileWidth / 2, -tileHeight / 2);
                t.scale((double)(rw * tileWidth) / up.getWidth(), (double)(rh * tileHeight) / up.getHeight());
                g.drawImage(up, t, this);
            }
        }
    }
}
